use std::{fmt::Display, str::FromStr, io::{stdin, Read}, collections::{BTreeMap, BTreeSet}};
use anyhow::{anyhow, Error, Result};

mod pollard;

#[derive(Debug, Clone, Copy, PartialEq, Eq, PartialOrd, Ord)]
enum Direction { L, R }

#[derive(Debug, Clone, Copy, PartialEq, Eq, PartialOrd, Ord)]
struct Node(u16);

impl FromStr for Node {
    type Err = Error;

    fn from_str(s: &str) -> Result<Self, Self::Err> {
        if !(s.len() == 3 && s.is_ascii()) { return Err(anyhow!("bad string {s}"))}
        let s = s.as_bytes();
        let x: u16 = (s[0] - b'A') as u16;
        let x: u16 = x*26 + (s[1] - b'A') as u16;
        let x: u16 = x*26 + (s[2] - b'A') as u16;
        Ok(Node(x))
    }
}

impl Display for Node {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        let c = (self.0 % 26) as u8 + b'A';
        let r: u16 = self.0 / 26;
        let b = (r % 26) as u8 + b'A';
        let a = (r / 26) as u8 + b'A';
        
        write!(f, "{}", std::str::from_utf8(&[a,b,c]).map_err(|_| std::fmt::Error)?)
    }
}

#[derive(Debug)]
struct ArrayMap(Vec<(Node,Node)>);

impl ArrayMap {
    fn get(&self, n: Node) -> &(Node,Node) {
        &self.0[n.0 as usize]
    }
}

impl FromIterator<(Node, (Node, Node))> for ArrayMap {
    fn from_iter<T: IntoIterator<Item = (Node, (Node, Node))>>(iter: T) -> Self {
        let mut map = vec![(Node(0), Node(0)); 26*26*26];
        for (s, pair) in iter.into_iter() {
            map[s.0 as usize] = pair;
        }
        ArrayMap(map)
    }
}


#[derive(Debug)]
struct Problem {
    map: BTreeMap<Node, (Node,Node)>,
    path: Vec<Direction>,
}

fn parse_entry(s: &str) -> Option<(&str, &str, &str)> {
    let (start, pair) = s.split_once(" = ")?;
    let (left, right) = pair
        .strip_prefix("(")?
        .strip_suffix(")")?
        .split_once(", ")?;
    Some((start, left, right))
}

impl FromStr for Problem {
    type Err = Error;

    fn from_str(s: &str) -> std::prelude::v1::Result<Self, Self::Err> {
        let mut lines = s.lines();
        let path = lines.next().ok_or(anyhow!("empty line"))?
            .chars()
            .map(|c| match c {
                'L' => Ok(Direction::L),
                'R' => Ok(Direction::R),
                c => Err(anyhow!("Invalid char {c}")),
            })
            .collect::<Result<Vec<_>>>()?;

        lines.next().filter(|l| l.is_empty()).ok_or(anyhow!("missing separator line"))?;

        let map = lines
            .map(|s| {
                let (start, left, right) = parse_entry(s).ok_or(anyhow!("Bad line {s}"))?;
                Ok((start.parse()?, (left.parse()?, right.parse()?)))
            })
            .collect::<Result<_>>()?;

        Ok(Self { map, path })
    }
}

impl Problem {

    fn follow(&self, start: Node) -> impl Iterator<Item=Node> + Clone + '_ {
        self.path.iter().cycle()
            .scan(start, |node, step| {
                let next = self.map.get(node)?;
                Some(match step { Direction::L => next.0, Direction::R => next.1 })
            })

    }

    fn count_path(&self, start: Node, goal: Node) -> usize {
        self.follow(start).take_while(|&n| n != goal).count()
    }

    /* 
    fn count_path(&self, mut start: Node, goal: Node) -> Result<usize> {
        let mut count = 0;
        let mut path = self.path.iter().cycle();
        while start != goal {
            count += 1;
            let next = self.map.get(&start).ok_or_else(|| anyhow!{"bad node {start}"})?;
            let step = path.next().expect("cycled iterator empty");
            start = match step { Direction::L => next.0, Direction::R => next.1 }
        }
        Ok(count)
    }
    */

    fn multi_count(&self) -> Result<usize> {
        let mut count = 0;
        let mut path = self.path.iter().cycle().copied();
        let mut starts: Vec<_> = self.map.keys().copied()
            .filter(|&Node(n)| n % 26 == 0)
            .collect();
        let cache: ArrayMap = self.map.iter()
            .map(|(s,p)| (*s, *p))
            .collect();

        while !starts.iter().copied().all(|Node(n)| n % 26 == 25) {
            count += 1;
            let step = path.next().expect("cycle iterator empty");
            
            for node in &mut starts {
                let next = cache.get(*node);
                *node = match step { Direction::L => next.0, Direction::R => next.1 }
            }
        }
        Ok(count)
    }
}

fn main() -> Result<()> {
    let mut data = String::new();
    stdin().lock().read_to_string(&mut data)?;

    let pb: Problem = data.parse()?;
    let start: Node = "AAA".parse()?;
    let stop: Node = "ZZZ".parse()?;

    let steps = pb.count_path(start, stop);
    println!("Step 1: {steps} steps.");

    let steps = pb.multi_count()?;
    println!("Step 2: {steps} steps.");

    Ok(())
}
